#include <stdlib.h>
#include <string.h>
#include <limits.h>

typedef unsigned char byte;

byte *pool, *pool_ptr, *helper, *helper_ptr;
size_t pool_size, left_space;
const byte SIZE_MARK_MAX = UCHAR_MAX >> 1;

void use_mj_allocator(size_t size)
{
	pool = pool_ptr = (byte *)malloc(size);
	helper = helper_ptr = (byte *)calloc(size, 1);
	pool_size = left_space = size;
}

void free_mj_allocator()
{
	free(pool);
	free(helper);
}

void set_size_mark(size_t size, byte *ptr)
{
	if (size <= SIZE_MARK_MAX)
	{
		*ptr = (size << 1) + 1;
	}
	else
	{
		*ptr = 2;
		*(size_t *)(ptr + 1) = size;
	}
}

size_t get_size_mark(const byte *ptr)
{
	if (*ptr & 1)
	{
		return *ptr >> 1;
	}
	return *(size_t *)(ptr + 1);
}

int search(size_t size)
{
	if (size > pool_size)
	{
		return 0;
	}
	byte *start_ptr = helper_ptr;
	if (left_space == 0)
	{
		goto no_left_space;
	}
	helper_ptr += left_space;
	left_space = 0;
	while (1)
	{
		if (helper_ptr == start_ptr)
		{
			return 0;
		}
	no_left_space:
		if (helper_ptr == helper + pool_size)
		{
			helper_ptr = helper;
			left_space = 0;
		}
		if (*helper_ptr)
		{
			left_space = 0;
			helper_ptr += get_size_mark(helper_ptr);
		}
		else
		{
			++left_space;
			++helper_ptr;
		}
		if (left_space == size)
		{
			while (1)
			{
				if (helper_ptr == helper + pool_size || *helper_ptr)
				{
					break;
				}
				++left_space;
				++helper_ptr;
			}
			break;
		}
	}
	helper_ptr -= left_space;
	pool_ptr = helper_ptr - helper + pool;
	return 1;
}

void *mj_malloc(size_t size)
{
	if (size > left_space)
	{
		if (!search(size))
		{
			return NULL;
		}
	}
	set_size_mark(size, helper_ptr);
	byte *ptr = pool_ptr;
	pool_ptr += size;
	helper_ptr += size;
	left_space -= size;
	return ptr;
}

void *mj_calloc(size_t len, size_t unit)
{
	size_t size = len * unit;
	void *ptr = mj_malloc(size);
	memset(ptr, 0, size);
	return ptr;
}

void mj_free(void *ptr)
{
	byte *temp_helper_ptr = (byte *)ptr - pool + helper;
	if (*temp_helper_ptr & 1)
	{
		*temp_helper_ptr = 0;
	}
	else
	{
		memset(temp_helper_ptr, 0, sizeof(size_t) + 1);
	}
}

size_t kb_to_b(size_t kb)
{
	return 1000 * kb;
}

size_t mb_to_b(size_t mb)
{
	return kb_to_b(1000 * mb);
}

size_t gb_to_b(size_t gb)
{
	return mb_to_b(1000 * gb);
}